/*
  This file is part of TALER
  Copyright (C) 2022 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3, or
  (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public
  License along with TALER; see the file COPYING.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file testing_api_cmd_delete_webhook.c
 * @brief command to test DELETE /webhooks/$ID
 * @author Priscilla HUANG
 */
#include "platform.h"
#include <taler/taler_exchange_service.h>
#include <taler/taler_testing_lib.h>
#include "taler_merchant_service.h"
#include "taler_merchant_testing_lib.h"


/**
 * State of a "DELETE /webhooks/$ID" CMD.
 */
struct DeleteWebhookState
{

  /**
   * Handle for a "DELETE webhook" request.
   */
  struct TALER_MERCHANT_WebhookDeleteHandle *wdh;

  /**
   * The interpreter state.
   */
  struct TALER_TESTING_Interpreter *is;

  /**
   * Base URL of the merchant serving the request.
   */
  const char *merchant_url;

  /**
   * ID of the webhook to run DELETE for.
   */
  const char *webhook_id;

  /**
   * Expected HTTP response code.
   */
  unsigned int http_status;

};


/**
 * Callback for a /delete/webhooks/$ID operation.
 *
 * @param cls closure for this function
 * @param hr response being processed
 */
static void
delete_webhook_cb (void *cls,
                   const struct TALER_MERCHANT_HttpResponse *hr)
{
  struct DeleteWebhookState *dis = cls;

  dis->wdh = NULL;
  if (dis->http_status != hr->http_status)
  {
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u (%d) to command %s\n",
                hr->http_status,
                (int) hr->ec,
                TALER_TESTING_interpreter_get_current_label (dis->is));
    TALER_TESTING_interpreter_fail (dis->is);
    return;
  }
  switch (hr->http_status)
  {
  case MHD_HTTP_NO_CONTENT:
    break;
  case MHD_HTTP_UNAUTHORIZED:
    break;
  case MHD_HTTP_NOT_FOUND:
    break;
  case MHD_HTTP_CONFLICT:
    break;
  default:
    GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
                "Unhandled HTTP status %u for DELETE webhook.\n",
                hr->http_status);
  }
  TALER_TESTING_interpreter_next (dis->is);
}


/**
 * Run the "DELETE webhook" CMD.
 *
 *
 * @param cls closure.
 * @param cmd command being run now.
 * @param is interpreter state.
 */
static void
delete_webhook_run (void *cls,
                    const struct TALER_TESTING_Command *cmd,
                    struct TALER_TESTING_Interpreter *is)
{
  struct DeleteWebhookState *dis = cls;

  dis->is = is;
  dis->wdh = TALER_MERCHANT_webhook_delete (is->ctx,
                                            dis->merchant_url,
                                            dis->webhook_id,
                                            &delete_webhook_cb,
                                            dis);
  GNUNET_assert (NULL != dis->wdh);
}


/**
 * Free the state of a "DELETE webhook" CMD, and possibly
 * cancel a pending operation thereof.
 *
 * @param cls closure.
 * @param cmd command being run.
 */
static void
delete_webhook_cleanup (void *cls,
                        const struct TALER_TESTING_Command *cmd)
{
  struct DeleteWebhookState *dis = cls;

  if (NULL != dis->wdh)
  {
    GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
                "DELETE /webhooks/$ID operation did not complete\n");
    TALER_MERCHANT_webhook_delete_cancel (dis->wdh);
  }
  GNUNET_free (dis);
}


struct TALER_TESTING_Command
TALER_TESTING_cmd_merchant_delete_webhook (const char *label,
                                           const char *merchant_url,
                                           const char *webhook_id,
                                           unsigned int http_status)
{
  struct DeleteWebhookState *dis;

  dis = GNUNET_new (struct DeleteWebhookState);
  dis->merchant_url = merchant_url;
  dis->webhook_id = webhook_id;
  dis->http_status = http_status;
  {
    struct TALER_TESTING_Command cmd = {
      .cls = dis,
      .label = label,
      .run = &delete_webhook_run,
      .cleanup = &delete_webhook_cleanup
    };

    return cmd;
  }
}


/* end of testing_api_cmd_delete_webhook.c */
